import osmnx as ox
import networkx as nx
import folium
import itertools
import pickle
import time, datetime
import os.path
import sys
import webbrowser

import random
 
import os
import glob


from shapely.geometry import LineString
from shapely.geometry import MultiLineString
from shapely.geometry import MultiPolygon
from shapely.geometry import Point
from shapely.geometry import Polygon
from shapely.ops import linemerge

import matplotlib.pyplot as plt
plt.switch_backend('agg')


ox.settings.log_console=True
ox.settings.use_cache=True

print('OSMnx version: ', ox.__version__)
print('NetworkX version: ',nx.__version__)
print('Python version: ',sys.version)

start_time = time.time()

# select between new and old algorithms
new_algo = True
#new_algo = False 

def elapsed_time(start):
    end_time = time.time()
    duration = round(end_time - start)
    timestamp = str(datetime.timedelta(seconds = duration))
    print(timestamp)

def graph_length(Q):
    length = ox.stats.edge_length_total(Q) #in meters
    length = length * 0.0006213712
    return length

def get_bounds(Q):
    QD = Q.to_directed()
    QD.graph['crs'] = ox.settings.default_crs    
    nodes, edges = ox.graph_to_gdfs(QD)
    tb = edges.total_bounds
    print( tb )
    sw = [tb[1],tb[0]]
    ne = [tb[3],tb[2]]
    return [sw, ne]

def save_png(Q, max_Q=None, bounds=None, index = 0):
    m = folium.Map()
    QD = Q.to_directed()
    QD.graph['crs'] = ox.settings.default_crs    
    nodes, edges = ox.graph_to_gdfs(QD)
    length = graph_length(Q)
    iteration_label = 'iteration '+ str(index) + ' ' + str(round(length)) + ' miles'
    ax = edges.plot(color='#0000ff', alpha=0.2, zorder=1, label = iteration_label)

    if max_Q != None:
        QD = max_Q.to_directed()
        QD.graph['crs'] = ox.settings.default_crs
        nodes, edges = ox.graph_to_gdfs(QD)
        max_length = graph_length(max_Q)
        max_label = 'current max: ' + str(round(max_length)) + ' miles'            
        edges.plot(ax=ax, color='#ff0000', alpha=1.0 , zorder=0, linestyle='dotted', label = max_label)
    if bounds != None:
        sw = bounds[0]
        ne = bounds[1]
        ax.set_xlim(sw[1], ne[1])
        ax.set_ylim(sw[0], ne[0])
    ax.legend(loc="lower left")
    ax.plot()
    plt.axis('off')
    plt.savefig('fig/img{}.png'.format(index), dpi=120)
    plt.close()


def report(Q, txt=''):
    print(len(Q.nodes()), ' nodes ', txt)
    print(len(Q.edges()), ' edges ', txt)
    if isinstance(Q,nx.MultiGraph):    
        length = ox.stats.edge_length_total(Q) #in meters
        length = length * 0.0006213712
        print(f'total length {length:.2f} miles ', txt)
    zeroes = 0    
    odds = 0
    evens = 0
    twos = 0
    ones = 0
    fours = 0
    four_pairs = 0
    parallels = 0
    for node in Q.nodes:
        d = Q.degree(node)
        if d == 0:
            zeroes += 1
        if d % 2 == 1:
            odds += 1
        else:
            evens += 1
        if d == 2:
            twos += 1
        if d == 1:
            ones += 1
        if d==4:
            fours += 1
            for neighbor in Q.neighbors(node):
                if Q.degree(neighbor)==4 and neighbor>node: # prevent double-count
                    four_pairs += 1
        if d>6:
            print('node ', node, 'has degree ', d)
        for neighbor in Q.neighbors(node):
            edge_count = Q.number_of_edges(node,neighbor)
            if edge_count > 1:
                parallels += edge_count - 1
                
    if zeroes > 0:            
        print('number of zero-degree nodes ' + txt + ' :', zeroes)            
    print('number of odd intersections ' + txt + ' :', odds)
    print('number of even intersections ' + txt + ' :', evens)
    print('number of 2-way intersections ' + txt + ' :', twos)
    print('number of 4-way intersections ' + txt + ' :', fours)
    print('number of trailhead/terminus ' + txt + ' :', ones)
    print('number of parallel edges ' + txt + ' :', parallels)
    if four_pairs > 0:
        print('number of 4-way adjacent pairs ' + txt + ' :', four_pairs)

        
    check_node_geometry(Q)
    print('')

def my_cmap(value):
    return value

def colorize_graph(Q):
    QR = Q.copy()
    for node, dat in Q.nodes(data=True):
        d = Q.degree(node)
        dat['color'] = d

        QR.add_node(node,**dat)                #replace node value

    for u, v, k, dat in Q.edges(keys=True,data=True):
        d = 'path'
        if 'highway' in dat:
            d = dat['highway']
        dat['color']=0.0
        if isinstance(d, str):
            if d=='path' or d == 'footway':
                dat['color']=1.0
            else:
                dat['color'] = 2.0
        QR.add_edge(u,v,key=k,**dat)
    QR.graph['color']=True
    return QR

def colorize_self_loops(Q,color=9.0):
    QR = colorize_graph(Q)
    for u,v,k,dat in QR.edges(keys=True,data=True):
        if u==v:
            QR[u][v][k]['color'] = color
            #dat['color']=color
            #QR.add_edge(u,v,key=k,**dat)
    return QR
    
                 
def colorize_edge_list(Q,edge_list,color=10.0):
     QR = colorize_graph(Q)
     for u,v in edge_list:
         dat = QR.nodes[u]
         dat['color'] = color
         QR.add_node(u,**dat)
         dat = QR.nodes[v]
         dat['color'] = color
         QR.add_node(v,**dat)
         if QR.has_edge(u,v):
             for k in QR[u][v]:
                 QR[u][v][k]['color']= color 
            # for v1,k in QR[u]:
             #    if v1==v:
              #       QR.edges[u,v,k]['color'] = color  # fix key
        #else if QR.has_edge(v,u):
        #    QR.edges[v,u,0]['color'] = 10.0
     return QR

def colorize_parallel_edges(Q):
    return colorize_edge_list(Q,parallel_edges(Q),color=9.0)

def parallel_edges(Q):
    ret_edges = set()
    for u,v in Q.edges():
        edge_count = Q.number_of_edges(u,v)
        if edge_count>1:
            ret_edges.add( ( min(u,v),max(u,v) ) )
    print('number of parallel edges: ',len(ret_edges))
    return ret_edges
    
def colorize_mismatched_nodes(Q,node_pairs,color=10.0):
    QR = colorize_graph(Q)
    matched_nodes = set()
    unmatched_nodes = set()
    for u,v in node_pairs:
        matched_nodes.add(u)
        matched_nodes.add(v)
    for node, dat in Q.nodes(data=True):
        if node not in matched_nodes:
            unmatched_nodes.add(node)
            dat['color'] = color
            QR.add_node(node,**dat) #update node attributes
            print('node ', node, ' not in matching')
    check_remainder_odd_nodes(J,unmatched_nodes)
    return QR

def check_node_geometry(Q):
    count = 0
    for node, dat in Q.nodes(data=True):
        if 'x' not in dat:
            count += 1
            #d = Q.degree(node)
            #print('node ', node, ' has no geometry attribute, degree ', d)t
    if count>0:
        print('number of nodes without geometry: ',count)


def check(Q,Q_original):
    remainder_nodes = get_remainder_odd_nodes(Q)
    #print('number of odd nodes remaining: ', len(remainder_nodes))
    check_remainder_odd_nodes(Q_original,remainder_nodes)
    for u,v,k,dic in Q.edges(data=True, keys=True):
        if 'node4_path' in dic:
            print('temporary path not deleted between: ',u,' and ',v, ' with key ', k)
        if not 'highway' in dic:
            print('suspicious edge: ',u,' and ',v, ' with key ', k)


def get_remainder_odd_nodes(Q):
    node_list = []
    for node in Q.nodes:
        if Q.degree(node) %2 ==1:
            node_list.append(node)
    return node_list

def check_remainder_odd_nodes(Q_original,node_list):
    remainder_odd_node_count = 0
    remainder_odd_node_4node_neighber_count = 0
    for node in node_list:
        if Q_original.degree(node)==4:
            remainder_odd_node_count +=1
            #print('remainder odd node ', node, 'was a 4node')
        else:
            for neighbor in Q_original.neighbors(node):
                if Q_original.degree(neighbor)==4:
                    #print('remainder odd node ',node,' has neighbor ', neighbor,' that is a 4node')
                    remainder_odd_node_4node_neighber_count +=1
    if remainder_odd_node_count > 0:
        print('remainder odd_node count: ', remainder_odd_node_count)
    if remainder_odd_node_4node_neighber_count > 0:
        print('remainder odd node with 4node neighbor count: ', remainder_odd_node_4node_neighber_count)
        

def check_self_loops(Q,txt=''):
    self_loops = 0
    self_loop_distance = 0
    max_self_loop_distance = 0
    for u,v,k,dic in Q.edges(keys=True,data=True):
        if u==v:
            d = dic['length']
            self_loops += 1
            self_loop_distance += d
            if d>max_self_loop_distance:
                max_self_loop_distance = d
    print('number of self-loops ' + txt + ' : ', self_loops )
    print(f'total self-loop distance {self_loop_distance*0.0006213712:.2f} miles ', txt)    
    print(f'maximum self-loop distance {max_self_loop_distance*0.0006213712:.2f} miles ', txt)
    print('')

def delete_self_loops(Q):
    self_loops = []
    for u,v,k,dic in Q.edges(keys=True,data=True):
        if u==v:
            self_loops.append( (u,v,k) )
    Q.remove_edges_from( self_loops)
    delete_ends(Q) # side-effect is that we left dangling ends
    
    
    
     
town_markers = [(34.29963, -108.13268, 'Pie Town'),
         (33.19593, -108.20834, 'Gila Hot Springs'),
         (32.77981, -108.27533, 'Silver City'),
         (33.37939, -108.90369, 'Alma/Glenwood'),
         (33.03022, -108.16864, 'Lake Roberts'),
         (32.85594, -107.98024, 'Mimbres'),       
         (32.91685, -107.70593, 'Kingston')]

# draw() changed after moving to Linux
def draw(Q,title='',show_nodes = False, boxes=None, not_boxes = None, markers=town_markers):
    start_latlng = (33.04876, -108.30680)
    m = folium.Map(location=[start_latlng[1],start_latlng[0]],
               zoom_start=12)
    # convert to MultiDiGraph for folium
    T = Q.copy()
    if not 'color' in T.graph:
        T = colorize_graph(T)
    QD = T.to_directed()
    QD.graph['crs'] = ox.settings.default_crs

    # our graphs may be HUGE, so simplify if not already done so
    if not "simplified" in QD.graph:
        QD = ox.simplify_graph(QD, endpoint_attrs=None, remove_rings=True, track_merged=True)
    
    nodes, edges = ox.graph_to_gdfs(QD)
#    m = edges.explore(color='#ff00ff', tiles="openstreetmap",tooltip='name')
    m = edges.explore(column='color',  cmap = 'tab10',legend=False, tiles="openstreetmap",tooltip='name')
    if show_nodes:    
#        nodes.explore(m=m, color="pink", marker_kwds={"radius": 4})    
        nodes.explore(m=m, column='color', cmap='tab10', legend=False,marker_kwds={"radius": 4})    

    if not boxes == None:
        for n,s,e,w in boxes:
            folium.Rectangle(bounds=[(n,e),(n,w),(s,e),(s,w)], color='#ff7800', fill=True, fill_color='#ffff00', fill_opacity=0.2).add_to(m)
    if not not_boxes == None:
        for n,s,e,w in not_boxes:
            folium.Rectangle(bounds=[(n,e),(n,w),(s,e),(s,w)], color='#00ff78', fill=False, fill_color='#00ffff', fill_opacity=0.2).add_to(m)
    if not markers == None:
        kw = {"prefix": "fa", "color": "green", "icon": "arrow-up"}
        icon = folium.Icon(angle=0, **kw)
        for lat,lon, txt in markers:
            #print(lat,lon,txt)
            folium.Marker(location=[lat,lon],popup=txt,
                          icon=folium.Icon(color='lightred',icon=''),
                          tooltip=txt).add_to(m)
            
    # include the option to switch map layers
   # folium.TileLayer('openstreetmap').add_to(m)
    folium.TileLayer('cartodbpositron').add_to(m)
   # folium.TileLayer('Stamen Terrain').add_to(m)
    folium.TileLayer(
        tiles='https://{s}.tile-cyclosm.openstreetmap.fr/cyclosm/{z}/{x}/{y}.png',
        name='cyclosm',
        attr='CyclOSM',
    ).add_to(m)
    folium.LayerControl().add_to(m)    

    title_html = '''
            <h3 align="center" style="font-size:16px"><b>{}</b></h3>
             '''.format(title)   
    m.get_root().html.add_child(folium.Element(title_html))
    file_name = "/home/sagebrush/biz/graph/map_" + title + ".html"
    m.save(file_name)
    #m.show_in_browser()
    webbrowser.open(file_name)
    input("Press Enter to continue...")
    print('*************', flush=True)
    print("", flush=True)    
        




#H = pickle.load(open('h_with_roadwalks_v2.pkl', 'rb'))
H = pickle.load(open('j_with_roadwalks_pruned.pkl', 'rb'))
report(H,'graph loaded from disk')






T = H.to_directed()
T = ox.simplification.simplify_graph(T, endpoint_attrs=None, remove_rings=False, track_merged=True)
#T = T.to_undirected()



#draw(T,'not-recommended trails removed', show_nodes=True)

#Note to self: We cannot draw until graph is simplified, unknown GDF problem

# we expect some disconnected subgraphs, so choose the largest
nodes = max(nx.connected_components(H), key=len)
J = nx.subgraph(H, nodes).copy()
report(J,'in largest subgraph')

# fix: this crashes
#draw(J,'simplified graph',show_nodes=True)


# Weight some trails so we get close to trail towns...
recommended = ['us 180', #glenwood/alma
                   'mineral creek trail',
                   'little whitewater trail', 
                   'arenas valley', #silver city
                   #'continental divide trail',
                   'apache creek trail', #lake roberts
                   'ladrone gulch trail', #kingston
                   'black range crest trail',
                   'ed jones road' #pie town
            
               ]
def weight_graph(Q):
    QR = Q.copy()
    for u, v, k, d in Q.edges(keys=True,data=True):
        if 'length' in d:
            wl=d['length']
            d['weighted_length']=wl
        else:
            d['weighted_length']=2 # we should not get here
        if 'name' in d:
            trail_name_list = d['name']
            trail_name = ''
            for t in trail_name_list:
                trail_name += t.lower()
            if any(e for e in recommended if e in trail_name):
                d['weighted_length'] = 1000
        QR.add_edge(u,v,key=k,**d)
    #Do further weights based on bridges
    bridge_list = list(nx.bridges(Q))
    for u,v in bridge_list:
        for k in Q[u][v]:
            QR[u][v][k]['weighted_length'] = 1000
        
    return QR

check(J,J)



# remove trailheads and termini
def delete_ends(Q):
    while True: # simulate do-while
        rn = []
        for node, neighbors in Q.adjacency():
            if len(neighbors) == 1:
                rn.append(node)
        if len(rn)>0:            
            Q.remove_nodes_from(rn)
        else:
            break

def cleanup_graph(Q):
    QR = nx.MultiGraph()
    bad_nodes = [240288183,2480249190]
    for u,v,k,d in Q.edges(keys=True, data=True):
        if u in bad_nodes or v in bad_nodes:
            continue
        QR.add_edge(u,v,k,**d)
    for node, d in Q.nodes(data=True):
        if node in bad_nodes:
            continue
        QR.add_node(node,**d)   #replace node value
    QR.graph['crs'] = ox.settings.default_crs    
    return QR
        

delete_ends(J)
##try:   
##    path = nx.shortest_path(J,240288183, 2480249139,weight='length')
##    if path != None and len(path)>0:
##        J.remove_edges_from(path)
##        print('removed suspicious edge at node 2480249139')
##    path = nx.shortest_path(J,2480249190, 2480249139,weight='length')
##    if path != None and len(path)>0:
##        J.remove_edges_from(path)
##        print('removed suspicious edge at node 2480249139')
##    J.remove_edge(2480249139,2480249190)
##except:
##    print('could not delete suspicious edge')

#J = cleanup_graph(J)
#print('neighbors of 2480249139: ',list(J.neighbors(2480249139)))
#print('degree of 2480249139: ', J.degree(2480249139))

def fix_covalent_4nodes(Q):
    QR = Q.copy()
    for u, v, d in Q.edges(data=True):
        if Q.degree(u) % 2 == 0 or Q.degree(v) % 2 == 0:
            edge_count = Q.number_of_edges(u,v)
            if edge_count==2:
                if QR.has_edge(u,v,1):
                    QR.remove_edge(u,v,1) #fix: find smallest edge
    return QR
 

T = J.to_directed()
T = ox.simplification.simplify_graph(T, remove_rings=True, track_merged=True,endpoint_attrs=None)
J = T.to_undirected()
nodes = max(nx.connected_components(J), key=len)
J = nx.subgraph(J, nodes).copy()

# ugly to put here, but we have one last final edit due to latest intelligence.
edges_to_remove = []
for u,v,k,d in J.edges(keys=True,data=True):
    if 'name' in d:
        trail_name = d['name'] # trail name will be a list here
        if 'Cooney Canyon/Mineral Creek Trail' in trail_name and 'McKean' not in trail_name:
                edges_to_remove.append( (u,v,k) )
                print('removed: ', trail_name)
J.remove_edges_from(edges_to_remove)
#draw(J,'upper mineral creek removed')



#J = colorize_graph(J)
J = weight_graph(J)  # we can only do this AFTER simplification

check(J,J)
delete_self_loops(J)
check_self_loops(J,'after simplification')

J = fix_covalent_4nodes(J)
report(J,'simplified graph max subgraph')

check_node_geometry(J)



#draw(J,'simplified graph', show_nodes=True)

#T = J.copy()
check_self_loops(J,'simplified graph')
#T = colorize_self_loops(T,7.0)
delete_self_loops(J)
check(J,J)
#check_self_loops(T,'removed self_loops')
#draw(T,'deleted self_loops',show_nodes=True)
#bridge_list = list(nx.bridges(T))
#print('length of bridges:', len(bridge_list))
#T = colorize_edge_list(T, bridge_list)
#draw(T,'bridges on simplified graph', show_nodes=True)


        
# get a graph of just the odd nodes, improve later to handle 4-way intersections

def odd_nodes_graph(Q):
    R = Q.copy()
    unwanted_edges = []
    for u, v, k in R.edges(keys=True):
        if R.degree(u) % 2 ==0 or R.degree(v) % 2 ==0:
            for k in R[u][v]:  #remove all edges between u and v, not just first one
                unwanted_edges.append( (u,v,k) )
    R.remove_edges_from( unwanted_edges )
    unwanted_nodes = []
    for node in R.nodes:
        if R.degree(node)==0:
            unwanted_nodes.append(node)
    R.remove_nodes_from( unwanted_nodes ) 
    return R
    

K = odd_nodes_graph(J)
report(K,'odd nodes graph')
check(K,J)



all_matchings = []
max_length = 0
def complete_3node_graph_old(Q,R):
    QR = R.copy()
    n4list = []
    for node, neighbors in Q.adjacency():        
        if len(neighbors) == 4:
            n4list.append( (node,neighbors) )
            nlist = []
            llen = []
            wlen = []
            for neighbor, edict in neighbors.items():
                w = min([d['weighted_length'] for d in edict.values()])
                l = min([d['length'] for d in edict.values()]) #fix
                nlist.append(neighbor)
                llen.append(l)
                wlen.append(w)
              #  for dic in edict.items():
                    #w = dic['length']
                    #nlist.append(neighbor)
                    #llen.append(w)
            # could use itertools?
            for i,j in [ (0,1), (1,2), (2,3), (3,0), (0,2),( 1,3) ]:
                QR.add_edge(nlist[i],nlist[j], weighted_length= wlen[i]+wlen[j],
                            length=llen[i]+llen[j]) 
    return n4list,QR



# ******* complete 3node graph ***************
# Replace 4nodes with temporary edges so that all nodes have odd degree
# Add a special property to the temporary edge dict so we can restore them to the original edges later.

def complete_3node_graph_new(Q,R): #original, 3node
    QR = R.copy()
    n4set = set()
    n4pairs = set()
    n4neighbors = set()

# get set of 4nodes
# For this next section, we could merge for loops for efficiency,
# but separate is clearer
    for node in Q:
        dg = Q.degree(node)

        # There shouldn't be any 2nodes in simplified graph, but there are.
        if dg==2:
            n2neighbor = []
            l2 = 0
            w2 = 0
            for u,v,k,d in Q.edges(node,keys=True,data=True,default=0):
                if u != node:
                    n2neighbor.append(u)
                elif v != node:
                    n2neighbor.append(v)
                l2 += d['length']
                w2 += d['weighted_length']
            if len(n2neighbor) ==2:
                QR.add_edge(n2neighbor[0],n2neighbor[1],weighted_length = w2,
                            length = l2,
                            node4_path= [ n2neighbor[0],node,n2neighbor[1] ] ) #fix: add geometry for draw()

        
        if dg>2 and dg%2==0:  # well, we had a 6 way intersection
            n4set.add(node)
    # get set of 4nodes with 4node neighbors
    for node in n4set:
        for neighbor in Q.neighbors(node):
            if (neighbor in n4set) and (neighbor != node):
                n4pairs.add( ( min(node,neighbor),max(node,neighbor) ) )
    # add bridging edges for n4set
    for node in n4set:
        list_edges = []
        for u,v,k,d in Q.edges(node,keys=True,data=True,default=0):
            # remove the edge case of 4node connected to a 4node
            if (u not in n4set) :    
    #        if (u not in n4set) and (u not in n4neighbors):    
                list_edges.append( (v,u,k,d) )
                n4neighbors.add(u)
            elif (v not in n4set) :
    #        elif (v not in n4set) and (v not in n4neighbors):
                list_edges.append( (u,v,k,d) )
                n4neighbors.add(v)
        for e1,e2 in itertools.combinations(list_edges,2):
            QR.add_edge(e1[1],e2[1],weighted_length = e1[3]['weighted_length']+e2[3]['weighted_length'],
                        length = e1[3]['length'] + e2[3]['length'],
                        node4_path=[ e1[1],node,e2[1] ]) # in add bridging edges for n4pairs, somewhat more complicated
##    for node,node2 in n4pairs:
##        list_edges = []
##        list_edges2 = []
##        for u,v,k,d in Q.edges(node,keys=True,data=True,default=0):
##        # remove the edge case of 4node connected to a 4node
##            if u not in n4set:    # first node in tuple is 4node
##                list_edges.append( (v,u,k,d) )
##            elif v not in n4set:
##                list_edges.append( (u,v,k,d) )
##        for u,v,k,d in Q.edges(node2,keys=True,data=True,default=0):
##        # remove the edge case of 4node connected to a 4node
##            if u not in n4set:    
##                list_edges2.append( (v,u,k,d) )
##            elif v not in n4set:
##                list_edges2.append( (u,v,k,d) )
##        d={}
##        d['weighted_length'] = Q.edges[node,node2,0]['weighted_length'] #fix: parallel edges for bridge
##        d['length'] = Q.edges[node,node2,0]['length'] #fix: parallel edges for bridge
##        if Q.number_of_edges(node,node2)!=1:
##            print('parallel edges between 4nodes ', node, ' ', node2)
##        bridge = (node,node2,0,d) #fix edge case multiple edges between node,node2
##        for e1 in list_edges:
##            for e2 in list_edges2:
##                QR.add_edge(e1[1],e2[1],weighted_length = e1[3]['weighted_length']+e2[3]['weighted_length']+bridge[3]['weighted_length'],
##                            length = e1[3]['length'] + e2[3]['length'] + bridge[3]['length'],
##                            node4_path=[ e1[1],node,node2,e2[1] ] ) #fix: add geometry for draw()
    print('number of 4node-pairs bridged: ', len(n4pairs))
    return QR

k_fix_4_list = []
def complete_3node_graph(Q,R): #original, 3node
    if new_algo:
        return complete_3node_graph_new(Q,R)
    else:
        n4list,QR = complete_3node_graph_old(Q,R)
        k_fix_4_list = n4list
        return QR


### The network will no longer be all odd, but matching should still work.
###Tried this patch to make all odd, but causes problems.
#####experimental patch
####add_edge = []
####for node, node2, d in K.edges(data='length'):
####    if K.degree(node)==4 and K.degree(node2)==4:
####        add_edge.append( (node,node2,d) )
####for u,v,d in add_edge:
####    K.add_edge(u,v,length=d)
##        
        
K = complete_3node_graph(J,K)
report(K,'in odd-nodes 4node bypass')
#draw(K, 'odd-nodes 4node bypass', show_nodes=True) #cannot draw unless we add geometries
# fix: find anything breaking odd-node



def remove_first_edges_from(Q,edges):
    QR = Q.copy()
    for u,v in edges:
        for k in range(9):
            if QR.has_edge(u,v,k):
                QR.remove_edge(u,v,k)
                break
    return QR

def find_shortest_edge(Q,u,v):
    min_k = 0
    min_length = -1
    for k in Q[u][v]:
        length = Q[u][v][k]['weighted_length']
        if length <= min_length or min_length < 0:
            min_length = length
            min_k = k
    return min_k

def find_longest_edge(Q,u,v):
    max_k = 0
    max_length = 0
    for k in Q[u][v]:
        length = Q[u][v][k]['weighted_length']
        if length > max_length:
            max_length = length
            max_k = k
    return max_k


def remove_shortest_edges_from(Q,edges):
    QR = Q.copy()
    for u,v in edges:
        QR.remove_edge(u,v, find_shortest_edge(Q,u,v))
    return QR

def graph_from_shortest_edges(Q):
    QR = nx.Graph()
    for u,v in Q.edges():
        k = find_shortest_edge(Q,u,v)
        wd = Q[u][v][k]['weighted_length']
        d = Q[u][v][k]['length']
        QR.add_edge(u,v,weighted_length = wd, length = d)
    return QR

            
def restore_4node_graph_old(Q,fix_4_list):
    for node, neighbors in fix_4_list:
       i = 0
       nlist = []
       vlist = []
       for neighbor, edict in neighbors.items():
           # for dic in edict:
           nlist.append(neighbor)
           vlist.append(edict[0]) //fix
       for i,j in [ (0,1), (1,2), (2,3), (3,0), (0,2),( 1,3) ]:               
            if Q.has_edge(nlist[i],nlist[j]):
                Q.remove_edge(nlist[i],nlist[j])
                if not Q.has_edge(node,nlist[i]):
                    Q.add_edge(node,nlist[i],**vlist[i]) #fix
                if not Q.has_edge(node,nlist[j]):                                  
                    Q.add_edge(node,nlist[j],**vlist[j])


def restore_4node_graph_new(Q,Q_original):
    QR = Q.copy()       
    for u,v,k,dic in Q.edges(data=True, keys=True):
        if 'node4_path' in dic:
            QR.remove_edge(u,v,key=k)
            path = dic['node4_path']
            # subgraph and compose do not seem to be working for me, so do explicitly
            for w,x in zip(path[:-1],path[1:]): #takes two nodes at a time
                if not QR.has_edge(w,x):
                    kr = find_longest_edge(Q_original,w,x) # fix restore remaining edge?
                    dat = Q_original.get_edge_data(w,x,kr)
                    QR.add_edge(w,x,kr,**dat)
##    n4pairs = set()
##    #go ahead and regenerate n4set and n4pairs, stay modular
##    for node in Q_original:
##        d = Q_original.degree(node)
##        if d>2 and d%2==0:  # well, we had a 6 way intersection
##            n4set.add(node)
##    # get set of 4nodes with 4node neighbors
##    for node in n4set:
##        for neighbor in Q_original.neighbors(node):
##            if (neighbor in n4set) and (neighbor != node):
##                n4pairs.add( ( min(node,neighbor),max(node,neighbor) ) )
##
##    print('n4set: ', n4set)
##    print('n4pairs: ', n4pairs)
##                     

##    n4pair_restore_count = 0
##    for u,v in n4pairs:
##        if QR.degree(u) %2==1 and QR.degree(v)%2==1:
##        #fix: handle case of deleted node
##            if QR.has_edge(u,v):  #fix: try else first
##                QR.remove_edge(u,v)
##                n4pair_restore_count +=1
##            else:
##                for k, dat in Q_original.edges[u][v]:
##                    QR.add_edge(u,v,k,dat)
##                    n4pair_restore_count +=1
##                    break
##    if n4pair_restore_count>0:
##        print('number of n4pairs adjusted: ', n4pair_restore_count)

    QR = fix_node_geometry(QR,Q_original)
    check_node_geometry(QR)
    QR = fix_remaining_odd_nodes(QR)
    return QR

def fix_node_geometry(Q,Q_original):
    restore_count = 0
    QR = Q.copy()
    for node in Q.nodes():  # add node attributes back in
        node_dic = Q.nodes[node]
        if 'x' not in node_dic:
            node_dic = Q_original.nodes[node]
            QR.add_node(node,**node_dic)
            restore_count += 1
   # if restore_count > 0:
    #    print('number of 4nodes restored: ', restore_count)
    return QR
    

def restore_4node_graph(Q,Q_original):
    QR = nx.MultiGraph()
    if new_algo:
        QR = restore_4node_graph_new(Q,Q_original)
    else:
        QR = Q.copy()
        restore_4node_graph_old(QR,k_fix_4_list)
    return QR
        
def fix_remaining_odd_nodes(Q): #fix: also look for adding edges
    QR = Q.copy()
    odd_nodes_fixed = set()
    delete_edges = []
    for u,v in Q.edges():
        if Q.degree(u) %2 ==1 and Q.degree(v) %2 ==1:
            if u not in odd_nodes_fixed and v not in odd_nodes_fixed:
                k = find_shortest_edge(Q,u,v)
                delete_edges.append( (u,v,k) )
                odd_nodes_fixed.add(u)
                odd_nodes_fixed.add(v)
    QR.remove_edges_from(delete_edges)
    return QR
        


    

all_matchings = []
max_length = 0

# delete png files from any previous run
files = glob.glob('fig/' + 'img*.png')
for f in files:
    os.remove(f)

bounds = get_bounds(J)
max_T2 = None
sub_max_T2 = None # used for iteration_algo = 3
max_length = 0
max_candidate_Graph = None
sub_max_length = 0 # iteration_algo 3
sub_max_candidate_Graph = None # iteration_algo 3


# fix: remove_edges_from on multigraph, might need custom



##def remove_shortest_edges_from(Q,node_pairs):
##    for u,v in node_pairs:
##        edge_count = Q.number_of_edges(node,neighbor)
##        min_k = 0
##        min_length = -1
##        for k, dic in Q.edges([u,v],keys=True, data=True):
##           current_length = dic['length']
##           if min_length <0:
##                min_k = k
##                min_length = current_length
##           elif current_length < min_length:
##                min_length = current_length
##                min_k = k
##        Q.remove_edge(u,v,min_k)

#matching = nx.max_weight_matching(nx.Graph(K),'length')
matching = nx.max_weight_matching(graph_from_shortest_edges(K),'weighted_length')
print('length of max_weight_matching = ', len(matching))
print('is perfect matching? ', nx.is_perfect_matching(K,matching))
T = K.copy()
colorize_mismatched_nodes(T,matching)
#T.remove_edges_from(matching) # fix: removed shortest edges from
T = remove_first_edges_from(T,matching)
T = restore_4node_graph(T,J)
report(T,'restored 4nodes max_weight_matching')
nodes = max(nx.connected_components(T), key=len)
T2 = nx.subgraph(T, nodes).copy()
report(T2,'restored 4nodes max_weight_matching max subgraph')
if nx.is_eulerian(T2):
    all_matchings.append(matching)
    report(T2, 'max_weight matching')
    print('length of successful matching: ', len(matching))
    print('max_weight matching is EULERIAN')
    length = graph_length(T2)
    all_matchings.append(matching)
    index = len(all_matchings)-1
    save_png(T2,max_T2,bounds,index)                                             
    if length > max_length:
        max_length = length
        max_T2 = T2.copy()

#matching = nx.min_weight_matching(nx.Graph(K),'length')
matching = nx.min_weight_matching(graph_from_shortest_edges(K),'weighted_length')
print('length of min_weight_matching = ', len(matching))
print('is perfect matching? ', nx.is_perfect_matching(K,matching))
colorize_mismatched_nodes(K,matching)
T = K.copy()
#T = colorize_edge_list(T,matching)
#draw(K,'min_weight_matching',show_nodes=True)
#T.remove_edges_from(matching)
#T = remove_first_edges_from(T,matching)
T = remove_shortest_edges_from(T,matching)
T=restore_4node_graph(T,J)
report(T,'restored 4nodes min weight matching')
T = colorize_graph(T)
#draw(T,'restored 4nodes min weight matching',show_nodes=True)
nodes = max(nx.connected_components(T), key=len)
T2 = nx.subgraph(T, nodes).copy()
report(T2, 'min_weight matching subgraph')
if nx.is_eulerian(T2):
    all_matchings.append(matching)
    report(T2, 'min_weight matching')
    print('length of successful matching: ', len(matching))
    print('min_weight matching is EULERIAN')
    length = graph_length(T2)
    all_matchings.append(matching)
    index = len(all_matchings)-1
    save_png(T2,max_T2,bounds,index)                                             
    if length > max_length:
        max_length = length
        max_T2 = T2.copy()

        
#matching = nx.min_weight_matching(nx.Graph(K)) #no weight
matching = nx.min_weight_matching(graph_from_shortest_edges(K)) #no weight
print('length of no-weight min_weight_matching = ', len(matching) )
print('is no-weight matching perfect? ', nx.is_perfect_matching(K,matching))
colorize_mismatched_nodes(K,matching)
T = K.copy()
#T.remove_edges_from(matching)
#T = remove_first_edges_from(T,matching)
T = remove_shortest_edges_from(T,matching)
T = restore_4node_graph(T,J)
nodes = max(nx.connected_components(T), key=len)
T2 = nx.subgraph(T, nodes).copy()
if nx.is_eulerian(T2):
    all_matchings.append(matching)
    print('length of successful matching: ', len(matching))
    print('no-weight matching is EULERIAN')
    length = graph_length(T2)
    index = len(all_matchings)-1
    save_png(T2,max_T2,bounds,index)
    if length > max_length:
        max_length = length
        max_T2 = T2.copy()
else:    
    print('no-weight matching is NOT eulerian')
report(T2, 'matching[1]')


def fuzz_min_weight_matching(Q):
    T = nx.MultiGraph()
    for u, v, k, d in Q.edges(data=True, keys=True):
        wl = d['weighted_length']
        l = d['length']
        wl = wl * (1 + random.random())
        T.add_edge(u,v,weighted_length=wl,length = l) #fix    
    return T, nx.min_weight_matching(graph_from_shortest_edges(T),weight='weighted_length')
def fuzz_max_weight_matching(Q):
    T = nx.MultiGraph()
    for u, v, k, d in Q.edges(data=True, keys=True):
        wl = d['weighted_length']
        wl = wl * (1 + random.random())
        l = d['length']
        T.add_edge(u,v,weighted_length=wl,length=l) #fix
    return T, nx.max_weight_matching(graph_from_shortest_edges(T),weight='weighted_length')
def random_weight_matching(Q):
    T = nx.MultiGraph()
    for u, v, k, d in Q.edges(data=True, keys=True):
        wl = random.random()
        l = d['length']
        T.add_edge(u,v,weighted_length=wl,length=l) #fix
    return T, nx.min_weight_matching(graph_from_shortest_edges(T),weight='weighted_length')



j=0
k=0
iteration_algo = 4



try:
    while j<10000 or k<1000:
    ##    if max_candidate_Graph == None:
    ##        candidate_Graph, matching = random_weight_matching(K)
    ##    else:
    ##        candidate_Graph, matching = random_weight_matching(max_candidate_Graph)

        if iteration_algo == 0 or (iteration_algo==4 and k>100):
            if max_candidate_Graph == None:
                candidate_Graph, matching = fuzz_min_weight_matching(K)
            else:
                candidate_Graph, matching = fuzz_min_weight_matching(max_candidate_Graph)
        elif iteration_algo == 1:
            if max_candidate_Graph == None:
                candidate_Graph, matching = fuzz_max_weight_matching(K)
            else:
                candidate_Graph, matching = fuzz_max_weight_matching(max_candidate_Graph)
        elif iteration_algo == 2 or (iteration_algo==4 and k<=100):
            if max_candidate_Graph == None:
                candidate_Graph, matching = random_weight_matching(K)
            else:
                candidate_Graph, matching = random_weight_matching(max_candidate_Graph)
        elif iteration_algo == 3:
            candidate_Graph, matching = random_weight_matching(K)

        elif iteration_algo == 3:
            if k % 200 == 0:
                sub_max_length = 0
                candidate_Graph, matching = random_weight_matching(K)
                print('starting new random candidate graph')
            else:
                if max_candidate_Graph == None:
                    candidate_Graph, matching = fuzz_min_weight_matching(K)
                else:
                    candidate_Graph, matching = fuzz_min_weight_matching(sub_max_candidate_Graph)

    #     colorize_mismatched_nodes(K,matching)
    #    colorize_parallel_edges(K)

        j+=1
        T = K.copy()
        #T.remove_edges_from(matching)
    #    T = remove_first_edges_from(T,matching)
        T = remove_shortest_edges_from(T,matching)
        T = restore_4node_graph(T,J)
        gen = (len(c) for c in sorted(nx.connected_components(T), key=len, reverse=True))
#        print('connected component lengths: ', *gen)
        nodes = max(nx.connected_components(T), key=len)
        T2 = nx.subgraph(T, nodes).copy()
        #report(T2, 'iteration graph')
        #draw(T,title='iteration',show_nodes = True)
    ##    if not nx.is_eulerian(T2): # not a perfect matching, so iterate once
    ##        T = T2.copy()
    ##        candidate_Graph, matching = fuzz_min_weight_matching(T)
    ####        candidate_Graph, matching = random_weight_matching(T)
    ##        T.remove_edges_from(matching)
    ##        T = restore_4node_graph(T,J)
    ##        nodes = max(nx.connected_components(T), key=len)
    ##        T2 = nx.subgraph(T, nodes).copy()
    ##        if nx.is_eulerian(T2):
    ##            print('SUCCESS IN FIXING non-eulerian matching')        


        if nx.is_eulerian(T2):
            k+=1
            all_matchings.append(matching)
            print('found EULERIAN ', k, ' on attempt ', j, )
            index = len(all_matchings)-1
            print('length of successful matching: ', len(matching))
            print('matching ', index,  ' is eulerian')
            #report(T2,'matching ' + str(index))
            length = graph_length(T2)
            save_png(T2,max_T2,bounds,index)
            if  iteration_algo == 3:
                if length > sub_max_length:
                    sub_max_length = length
                    sub_max_T2 = T2.copy()
                    sub_max_candidate_Graph = candidate_Graph.copy()
            if length > max_length:
                max_length = length
                max_T2 = T2.copy()
                max_candidate_Graph = candidate_Graph.copy()
                pickle.dump(max_T2,open('euler_graph.pkl', 'wb'))

            if iteration_algo ==3:
                print('current sub max length: ', round(sub_max_length))
            print('current max length: ',round(max_length),'\n')
       # else:
       #     colorize_mismatched_nodes(T2,matching)

except KeyboardInterrupt:
    print('\nProgram terminated by user.')
                                                
    
elapsed_time(start_time)
print('finished finding some possible matchings')
print('max length = ', max_length, ' miles')


pickle.dump(all_matchings,open('all_matchings_with_roadwalks.pkl', 'wb'))
#pickle.dump(fix_4_list,open('fix_4_list_with_roadwalks.pkl', 'wb'))
pickle.dump(K,open('k_maximum_with_roadwalks.pkl', 'wb'))
pickle.dump(J,open('j_maximum_with_roadwalks.pkl', 'wb'))
pickle.dump(max_T2,open('maxt2_with_roadwalks_' + str(round(max_length)) + '.pkl', 'wb'))
draw(max_T2,'final candidate ' + str(round(max_length)))


elapsed_time(start_time)

print('finished script ' + os.path.basename(__file__))






